/* psdslayer
 * Copyright 2009 Jetro Lauha
 *
 * This is simply the example code for using DrPetter's PSDImage class
 * combined with some options and stb_image + zlib + libpng for saving
 * each layer as a separate image file.
 *
 * My home page / blog URL:
 *   http://jet.ro
 *
 * Original URL for PSDImage release:
 *   http://www.ludumdare.com/compo/2008/04/12/psdimage-class/
 * URL for stb_image:
 *   http://nothings.org/stb_image.c
 *
 *
 * This software is provided 'as-is', without any express or implied
 * warranty.  In no event will the authors be held liable for any damages
 * arising from the use of this software.
 *
 * Permission is granted to anyone to use this software for any purpose,
 * including commercial applications, and to alter it and redistribute it
 * freely, subject to the following restrictions:
 *
 * 1. The origin of this software must not be misrepresented; you must not
 *    claim that you wrote the original software. If you use this software
 *    in a product, an acknowledgment in the product documentation would be
 *    appreciated but is not required.
 * 2. Altered source versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 * 3. This notice may not be removed or altered from any source distribution.
 *
 * $Id$
 * $Revision$
 */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "psdimage.h"
#include "png.h"

#define STBI_HEADER_FILE_ONLY
#include "stb_image.c"


void png_error_fn(png_structp png_ptr, png_const_charp msg)
{
    fprintf(stderr, "libpng ERROR: %s\n", msg);
    exit(EXIT_FAILURE);
}

void png_warn_fn(png_structp png_ptr, png_const_charp msg)
{
    fprintf(stderr, "libpng warning: %s\n", msg);
}


bool write_png(const char *resultFile, int width, int height, unsigned int *imagedata)
{
    FILE *fp = fopen(resultFile, "wb");
    if (fp == 0)
        return false;

    png_structp png_ptr = 0;
    png_infop info_ptr = 0;

    png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, 0, png_error_fn, png_warn_fn);
    if (png_ptr == 0)
    {
        fclose(fp);
        return false;
    }

    info_ptr = png_create_info_struct(png_ptr);
    if (info_ptr == 0)
    {
        fclose(fp);
        return false;
    }

    png_init_io(png_ptr, fp);

    png_set_IHDR(png_ptr, info_ptr, width, height, 8, PNG_COLOR_TYPE_RGB_ALPHA, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

    png_write_info(png_ptr, info_ptr);

    //png_set_bgr(png_ptr); // little-endian

    for (int y = 0; y < height; ++y)
    {
        png_write_row(png_ptr, (png_bytep)imagedata);
        imagedata += width;
    }

    png_write_end(png_ptr, info_ptr);

    png_destroy_write_struct(&png_ptr, &info_ptr);

    fclose(fp);

    return true;
}


int main(int argc, char *argv[])
{
    bool layernum = false;
    bool metatofn = false;
    char *prefix = "";
    bool usetga = false;

    printf("psdslayer v1.1\n");

    while (argc >= 2 && argv[1][0] == '-')
    {
        if (strcmp(argv[1], "-metatofn") == 0)
        {
            metatofn = true;
            --argc;
            ++argv;
        }
        else if (strcmp(argv[1], "-layernum") == 0)
        {
            layernum = true;
            --argc;
            ++argv;
        }
        else if (strcmp(argv[1], "-tga") == 0)
        {
            usetga = true;
            --argc;
            ++argv;
        }
    }

    if (argc < 2)
    {
        printf("Usage: psdslayer [-metatofn] [-layernum] [-tga] input.psd [prefix]\n\n");
        printf("Saves each layer as layername.png file with optional prefix.\n");
        printf("PSD file should be saved in \"maximized compatibility mode\" and\n");
        printf("without any special features such as layer groups.\n\n");
        printf("Warning: Destination files are overwritten!\n\n");
        printf("    -metatofn  use complex output name with prefix, layer\n");
        printf("               number, org. XY offset and layer name.\n\n");
        printf("    -layernum  add layer number to output filenames\n");
        printf("               (used without -metatofn)\n");
        printf("    -tga       save to .tga files instead of .png\n\n");
        return EXIT_SUCCESS;
    }

    if (argc >= 3)
        prefix = argv[2];

    PSDImage *img = new PSDImage();
    if (!img->Load(argv[1])) // load image file and store image data for all layers
    {
        printf("Error loading %s\n", argv[1]);
        return EXIT_FAILURE;
    }

    for (int i = 0; i < img->NumLayers(); ++i) // iterate over all layers (if desired)
    {
        // dimensions of this layer (outer boundaries of image content)
        int width;
        int height;
        // upper-left coordinate of bounding box, use this to place the layer accurately within the image
        int offset_x;
        int offset_y;
        DWORD *imagedata; // pixel array in ARGB format, need not be allocated (will point to stored image data)
        char name[33] = { 0 };
        memcpy(name, img->GetLayer(i)->name, 32);

        imagedata = img->GetLayerData(i, width, height, offset_x, offset_y); // get pointer to layer image, and info

        if (imagedata == NULL)
            printf("Layer %i: FAILED (%s)\n", i, name);
        else
            printf("Layer %i: \"%s\", width:%d, height:%d, x:%d, y:%d%s\n", i, name, width, height, offset_x, offset_y,
                   img->LayerHidden(i) ? ", HIDDEN" : "");

        //if (img->LayerHidden(i))
        //    printf("layer %i (%s) is hidden, but can be used anyway", i, name);

        
        if (width > 0 && height > 0)
        {
            int offs = 0;
            for (int y = 0; y < height; ++y)
            {
                for (int x = 0; x < width; ++x)
                {
                    char tmp, *pb = (char *)&imagedata[offs++];
                    // swap RB
                    tmp = pb[0];
                    pb[0] = pb[2];
                    pb[2] = tmp;
                }
            }
            for (int a = 0; a < 32 && name[a]; ++a)
            {
                int ch = name[a];
                if (!isalnum(ch) && ch != '-' && ch != '_')
                    name[a] = '_';
            }
            char resultFile[256];
            const char *ext = usetga ? "tga" : "png";
            if (metatofn)
                sprintf(resultFile, "%s%d-x%d_y%d_%s.%s", prefix, i, offset_x, offset_y, name, ext);
            else if (layernum)
                sprintf(resultFile, "%s%d-%s.%s", prefix, i, name, ext);
            else
                sprintf(resultFile, "%s%s.%s", prefix, name, ext);

            bool success;
            if (usetga)
            {
                int res = stbi_write_tga(resultFile, width, height, 4, imagedata);
                success = res != 0;
            }
            else
            {
                success = write_png(resultFile, width, height, (unsigned int *)imagedata);
            }
            if (!success)
            {
                printf("Error saving layer %i (%s)\n", i, name);
            }
        }

    }

    delete img; // note that this will free all the image data, so any pointers you might have to layers will be invalid

    // (copy the layer image data if you want to keep it after the PSD object has been destroyed)

    // it's safe to load a new PSD image using the same object, but all previous layer data will of course be lost


    return EXIT_SUCCESS;
}
